package mon.battle

import scala.util.Random

import mon.stat._
import mon.stat.Statistic._
import mon.util.Fraction

class BattleEngine(val player : Party, val enemy : Party)(implicit val rng : Random) {
  
  def playTurn() = {
    val playerMove = player.pollAction
    val playerTarget = player.pollTarget(playerMove, player.lead, player, enemy)
    val enemyMove = enemy.pollAction
    val enemyTarget = enemy.pollTarget(enemyMove, enemy.lead, enemy, player)
    val actions = Seq(Action(player.lead, playerMove, playerTarget), Action(enemy.lead, enemyMove, enemyTarget))
    val queue = rng.shuffle(actions).sorted // Shuffle to randomize in the event of a tie
    queue.foreach(useMove)
    val eotQueue = Seq(player.lead, enemy.lead).sortBy(_(Speed))
    eotQueue.foreach(mon => mon.status.map(_.onResidual.map(_(mon))))
    println(s"${player.lead}(${player.lead.hp}/${player.lead(Hp)})")
    println(s"${enemy.lead}(${enemy.lead.hp}/${enemy.lead(Hp)})")
  }
  
  def useMove(action : Action) = {
    val user = action.user
    val move = action.move
    val target = action.target
    println(s"$user used $move.")
    if (attackRoll(user, move, target)) {
      if (move.pow > 0 || move.powCallback != null) {
        val dmg = damageRoll(user, move, target)
        target.takeDamage(dmg)
        println(s"$target takes $dmg damage!")
      }
      applyBoosts(target, move.boosts)
      // TODO : Secondary effects
      if (move.selfEffect != null) {
        applyBoosts(user, move.selfEffect.boosts)
      }
      if (move.secondary != null && rng.chance(move.secondary.chance.%%)) {
        applyBoosts(target, move.secondary.boosts)
        if (move.secondary.status != null && target.status == None) {
          // apply status
          target.status = Some(move.secondary.status)
          move.secondary.status.onStart.map(_(target))
        }
      }

      // TODO : Support moves
      // TODO : Multiparty
    } else {
      println("Missed!")
    }
  }
  
  def attackRoll(user : Monster, move : Move, target : Monster) = {
    if (move.accuracy == 0) {
      true
    } else {
      val acc = user.boosts.getOrElse(Accuracy, 0) - target.boosts.getOrElse(Evasion, 0)
      val mod = if (acc > 0) Fraction(3 + acc, 3) else Fraction(3, 3 - acc)
      val chance = move.accuracy.%% * mod
      rng.chance(chance)
    }
  }
  
  def damageRoll(user : Monster, move : Move, target : Monster) = {
    val atkStat = if (move.mvType == MoveType.Physical) PAtk else MAtk
    val defStat = if (move.mvType == MoveType.Physical) PDef else MDef
    // TODO : Fixed damage
    val pow = if (move.powCallback != null) move.powCallback(user, target) else move.pow
    val baseDmg = (2 * user.level / 5 + 2) * pow * user(atkStat) / (target(defStat) * 50) + 2
    val multiplier = target.effectiveness(move.element)
    if (multiplier > 1.0) {
      println("It's super effective!")
    } else if (multiplier < 1.0) {
      println("It's not very effective.")
    }
    val maxDmg = (baseDmg * multiplier).toInt
    val minDmg = (17 \ 20) * maxDmg
    rng.nextInt(minDmg, maxDmg)
  }
  
  def applyBoosts(target : Monster, boosts : Map[Stat, Int]) {
    boosts.foreach{case (s, b) => {
      target.applyBoost(s, b)
      if (b > 0) {
        println(s"$target's $s rose!")
      } else if (b < 0) {
        println(s"$target's $s fell!")
      }
    }}
  }
}